/*
** Copyright (C) 1989 by Jef Poskanzer.
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/

/* Global constants and macros */
#define BITS 12
#define HSIZE 5003
#define MAXCODE(n_bits) (((int) 1 << (n_bits)) - 1)
#define HashTabOf(i) htab[i]
#define CodeTabOf(i) codetab[i]

#include <stdio.h>

/* Global variables */
extern FILE *Fd;

static int init_bits = 9;
static int clear_flg = 0;
static int maxcode = 0;
static int clear_code = 0;
static int eof_code = 0;
static int free_ent = 0;
static int a_count = 0;
static int hsize = HSIZE;
static long htab[HSIZE];
static unsigned short codetab[HSIZE];
static int n_bits = 0;
static int maxmaxcode = (int)1 << BITS; 
static int maxbits = BITS;
static unsigned long cur_accum = 0;
static char accum[ 256 ];
static int cur_bits = 0;
static unsigned long masks[] = 
   { 0x0000, 0x0001, 0x0003, 0x0007, 0x000F,
     0x001F, 0x003F, 0x007F, 0x00FF,
     0x01FF, 0x03FF, 0x07FF, 0x0FFF,
     0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF };

/* Flush the character buffer */
/* -------------------------- */
void flush_char()
{
  if( a_count > 0 ) 
    {
      fputc( a_count, Fd );
      fwrite( accum, 1, a_count, Fd );
      a_count = 0;
    }
}

/* Output a character to the buffer */
/* -------------------------------- */
void char_out( int c )
{
  accum[ a_count++ ] = c;
  if( a_count >= 254 )
    flush_char();
}

/* Output the compressed data */
/* -------------------------- */
void output( int code )
{
  cur_accum &= masks[ cur_bits ];
  if( cur_bits > 0 )
    cur_accum |= ((long)code << cur_bits);
  else
    cur_accum = code;
  cur_bits += n_bits;
  
  while( cur_bits >= 8 ) 
    {
      char_out( (unsigned int)(cur_accum & 0xff) );
      cur_accum >>= 8;
      cur_bits -= 8;
    }
  
  if ( free_ent > maxcode || clear_flg ) 
    {
      if( clear_flg ) 
	{
	  maxcode = MAXCODE (n_bits = init_bits);
	  clear_flg = 0;
	} 
      else 
	{
	  ++n_bits;
	  if ( n_bits == maxbits )
            maxcode = maxmaxcode;
	  else
            maxcode = MAXCODE(n_bits);
	}
    }
  
  if( code == eof_code ) 
    {
      while( cur_bits > 0 ) 
	{
	  char_out( (unsigned int)(cur_accum & 0xff) );
	  cur_accum >>= 8;
	  cur_bits -= 8;
	}
      flush_char();
      fflush( Fd );
      if( ferror( Fd ) )
	fprintf (stderr,"Error writing output file");
    }
}

/* Reset the hash code table */
/* ------------------------- */
void cl_hash()  
   {
   register long *htab_p = htab+hsize;
   register long i;
   register long m1 = -1;
   i = hsize - 16;
   do {  
      *(htab_p-16) = m1;
      *(htab_p-15) = m1;
      *(htab_p-14) = m1;
      *(htab_p-13) = m1;
      *(htab_p-12) = m1;
      *(htab_p-11) = m1;
      *(htab_p-10) = m1;
      *(htab_p-9) = m1;
      *(htab_p-8) = m1;
      *(htab_p-7) = m1;
      *(htab_p-6) = m1;
      *(htab_p-5) = m1;
      *(htab_p-4) = m1;
      *(htab_p-3) = m1;
      *(htab_p-2) = m1;
      *(htab_p-1) = m1;
      htab_p -= 16;
      } while ((i -= 16) >= 0);

   for ( i += 16; i > 0; --i )
      *--htab_p = m1;
   }

/* Clear out the hash table */
/* ------------------------ */
void cl_block () 
   {
   cl_hash ();
   free_ent = clear_code + 2;
   clear_flg = 1;
   output( (int)clear_code );
   }

/* Data compression function */
/* ------------------------- */
void compress(unsigned char *Data, int Count)
{
  long fcode;
  int i = 0;
  int ent;
  int data;
  int disp;
  int hsize_reg;
  int hshift;
  int Index;
  

  init_bits = 9;
  maxmaxcode = (int)1 << BITS; 
  maxbits = BITS;
  cur_accum = 0;
  cur_bits = 0;

  /* Initialize variables */
  clear_flg = 0;
  n_bits = init_bits;
  maxcode = MAXCODE(init_bits);
  clear_code = (1 << (init_bits - 1));
  eof_code = clear_code + 1;
  free_ent = clear_code + 2;
  a_count = 0;
  
  /* Set hash code range bound */
  hshift = 0;
  for ( fcode = (long) hsize;  fcode < 65536L; fcode *= 2L )
    ++hshift;
  hshift = 8 - hshift;                
  
  /* Clear hash table */
  hsize_reg = hsize;
  cl_hash();            
  
  /* Output compressed data */
  output( (int)clear_code );
  Index = 0;
  ent = (int)(Data[Index++]);
  while (Index<Count)
    {
      data = (int)(Data[Index++]);
      fcode = (long) (((long) data << maxbits) + ent);
      i = (((int)data << hshift) ^ ent);
      if ( HashTabOf (i) == fcode ) 
	{
	  ent = CodeTabOf (i);
	  continue;
	} 
      else if ( (long)HashTabOf (i) < 0 ) 
	goto nomatch;
      disp = hsize_reg - i;           /* secondary hash (after G. Knott) */
      if ( i == 0 )
	disp = 1;

    probe:

      if ( (i -= disp) < 0 )
	i += hsize_reg;
      
      if ( HashTabOf (i) == fcode ) 
	{
          ent = CodeTabOf (i);
          continue;
	}
      if ( (long)HashTabOf (i) > 0 )
	goto probe;

    nomatch:

      output ( (int) ent );
      ent = data;
      if ( free_ent < maxmaxcode ) 
	{ 
	  CodeTabOf (i) = free_ent++; /* code -> hashtable */
	  HashTabOf (i) = fcode;
	} 
      else
	cl_block();
    }
  
  /* Put out the final code */
  output( (int)ent );
  output( (int) eof_code );
}

