#include "eigen.hh"
#include "gif.hh"

#include <iostream.h>
#include <stdio.h>

/* #define READ_RAW 1 */

void floatbyteswap(float * x, float *y)
{
    ((char *) (y))[3] = ((char *) (x))[0];
    ((char *) (y))[0] = ((char *) (x))[3];
    ((char *) (y))[2] = ((char *) (x))[1];
    ((char *) (y))[1] = ((char *) (x))[2]; 
}

eigen::eigen(char	*path)
{
  path_ = new char[strlen(path)+1]; 	// Copy the name for later
  strcpy(path_, path);
  
  get_info();				// Get info from path/README file
  images_ = new intensity*[count_]; 	// allocate space for entire array

  valid_ = true;
  info();
}

void
eigen::get_info()
{
  
  char		pathname[200];
  char		data_line[1000];
  FILE		*fp;

  sprintf(pathname, "%s/%s", path_, "README");
  fp=fopen(pathname, "r");

  count_ = 0;

  if (fp == NULL)
    {
      cerr << pathname << " is not a valid path for an EIGEN-BASE" << endl;
      width_ = height_ = count_ = size_ = 0;
      return;
    }

  while (!feof(fp))
    {
      fscanf(fp, "%s", data_line);

      if (strstr(data_line, "data_files"))
	fscanf(fp, "%d", &count_);
      
      if (strstr(data_line,"dimension"))
	fscanf(fp,"%d %d", &height_, &width_);
    }

  size_ = height_ * width_;
  fclose(fp);
}


eigen::~eigen()
{
  delete path_;

  if (valid_)
    delete images_;
}

void
eigen::load()
{
  float		*data = new float[size_], *fp;
  char		pathname[200];  
  FILE		*FP;
  int		i;

  sprintf(pathname, "%s/%s.eig", path_, "data");
  
  FP=fopen(pathname, "r");
  
  if (FP == NULL)
    {
      cerr << pathname << " is not a valid path for an EIGEN-BASE" << endl;
      width_ = height_ = count_ = size_ = 0;
      return;
    }

  eigenvalues_= new float[count_];

  for ( i=count_ , fp=eigenvalues_; i; --i, fp++)
    fscanf(FP, "%f", fp);
}

void
eigen::load(int		index)
{
  float		*data = new float[size_], *fp;
  char		pathname[200];  
  FILE		*FP;
  int		i;

#ifdef READ_RAW
  sprintf(pathname, "%s/%s%d", path_, "data", index);
#else
  sprintf(pathname, "%s/%s%d.2", path_, "data", index);
#endif
  
  FP=fopen(pathname, "r");
  
  if (FP == NULL)
    {
      cerr << pathname << " is not a valid path for an EIGEN-BASE" << endl;
      width_ = height_ = count_ = size_ = 0;
      return;
    }

#ifdef READ_RAW
  fread(data, sizeof(float), size_, FP);

  // Now convert from whatever endian format the data is written in

   if (strcmp(getenv("FLOAT_SWAP"),"on") == 0) {
     int i,j;
     float f0;

     for (i=0; i<height_; i++) {
       for (j=0; j<width_; j++) {
	 floatbyteswap(data+i*width_+j, &f0);
	 data[i*width_+j] = f0;
       }
     }
   }

#else
  for ( i=size_, fp=data; i; --i, fp++)
    fscanf(FP, "%f", fp);
#endif

  images_[index] = new intensity(width_, height_, data);
//  images_[index]->normalize();
}

void				// face space metrics
eigen::d_f_s(intensity		&inten, int dimensionality)
{
  // clip the search area
  int		width = inten.width()-width_; 
  int		height = inten.height()-height_;

  register int	i,j,k;

  float		dummy, difs, dffs, energy, sp[20];

  // allocate a new distance from face space figure
  dfs_ = new intensity(width, height);

  for (  i=0; i < height; i++)
    {
      for (j=0; j < width; j++)
	{
	  
	  intensity		subimage( width_, height_, inten, j, i);
        // URS do normalisation here (unitmapping) 
          subimage.umap();
	  
	  subimage -= images_[0];		// subtract the mean image
	  energy=subimage.energy(); 	// total energy for face
	 
          dffs=energy; 
          difs=0; 
	  for (k=1; k <=  dimensionality; k++)
          {
	    sp[k]=subimage.project( images_[k]);
	    dummy=sp[k]*sp[k];
	    dffs -= dummy;
	    difs += dummy/eigenvalues_[k-1];
          }
	  
	  dfs_->set(j, i, difs+dffs/eigenvalues_[dimensionality]);
//	  cout << toten << " , ";
	}
//     cout << endl;
    }
//  dfs_->set(0,0, 1);

//  dfs_->normalize();
}

