#include "intensity.hh"
#include <math.h>
#define  MAX_CHARS 256

intensity::intensity(int	width, int	height, float		*data)
{
  aux_init(width, height, data);
  info();
}

intensity::intensity(int	width, int	height, unsigned char	*data)
{

  aux_init( width, height, NULL);
  info();
  aux2(data);
}

intensity::intensity(gif	&GIF)
{

  aux_init( GIF.width(), GIF.height(), NULL);
  info();
  aux2( GIF.data());

// URS: don't normalise here 
/*   normalize(); */
}

// URS: create intensity directly from file 
// implementation for P5 pgm so far (other image formats can be included here
// get pieces from Baback's 'io' code
intensity::intensity(char * filename)
{
    FILE *fp;
    register int i,j;
    int n;
    char str[10], line[MAX_CHARS];
    unsigned char *row;
    int height, width;
    float * image;
    int char_size = sizeof(char);

    if ((fp = fopen(filename, "r")) == NULL) 
    {
       fprintf(stderr,"ERROR: read_PGM could not open file %s\n\n",filename);
       exit(1);
    }
    /* --- parse image header ---- */

   fgets(line, MAX_CHARS, fp);
   sscanf(line, "%s", str);
   if (strncmp(str, "P5", 2)) {
      fprintf(stderr,"ERROR: Routine read_PGM found non-PGMRaw file.");
      exit(1);
   }
   while (fgets(line, MAX_CHARS, fp))
     if (strncmp(line, "#", 1)) {
       sscanf(line, "%d %d", &width, &height);
      break;
    }
    while (fgets(line, MAX_CHARS, fp))
      if (strncmp(line, "#", 1)) {
        sscanf(line, "%d", &n);
      break;    }

  /* --- allocate space ---- */

  valid_ = true;
  width_ = width;
  height_ = height;
  size_ = height_ * width_ ;
  image = new float[size_];
  n_image_ = new float * [height_];
  row = new unsigned char [width_];
 
  for (i=0;i< height;i++) 
  {
      n_image_[i]= &image[i*width_];
      fread((char *)row, char_size, width_, fp);
      for (j=0;j< width_ ;j++)
         n_image_[i][j] = (float) row[j];
  }

  delete row;
  fclose(fp);
}

intensity::intensity(intensity	&inten)
{
  aux_init( inten.width(), inten.height(), NULL);
  info();

  memcpy(data(), inten.data(), size_ * sizeof(float));

}

intensity::intensity(int		width, 	int		height,
		     intensity		&inten, 
		     int		st_x, int	st_y)
{
  register int		i,j;


  aux_init(width, height, NULL);
  info();
  
  for (i=st_y,j=0; j < height; i++,j++)
  {
     float * p=inten.n_image_[i];
     p+=st_x;
     memcpy(n_image_[j], p, width * sizeof(float));
  }
}

intensity::~intensity()
{

  if (!valid_)
    return;

  delete *n_image_;
  delete n_image_;
  valid_ = false;
}

void
intensity::aux2(unsigned char	*uc)
{
  float			*image;
  unsigned char		 max, *sp;
  float			flMax;

  sp=uc;
  max = 0;
  for(int i=size_; i ; --i, uc++)
    if (*uc > max)
      max = *uc;

  image = data();
  flMax = max;

  uc=sp;
  for(i=size_; i ; --i, image++, uc++)
      *image = (float) (*uc) / flMax;
}

void
intensity::aux_init(int		width, int		height, float	*data)
{
  float		*image;

  valid_ = true;
  
  width_  = width;
  height_ = height;
  size_   = height_ * width_;		// Assign the size 

  n_image_ = new float * [ height_]; 	// allocate space for array

  if (data == NULL)
    image=new float[size_]; 		// allocate space for data
  else
    image=data;


  for(int i=0; i < height_; i++)
     n_image_[i] = &image[ i*width_ ];
}

intensity &
intensity::operator+(float	RHS) 
{
  float		*image;

  image = data();

  for (int i=size_; i; --i, image++)
    *image += RHS;

  return *this;
}

intensity &
intensity::operator-(float	RHS) 
{
  float		*image;

  image = data();

  for (int i=size_; i; --i, image++)
    *image -= RHS;

  return *this;
}

intensity &
intensity::operator-=(float	RHS) 
{
  float		*image;

  image = data();

  for (int i=size_ ; i ; --i, image++)
    *image -= RHS;

  return *this;
}

intensity &
intensity::operator-=(intensity	*RHS) 
{
  float		*image;
  float		*image2;

  image = data();
  image2 = RHS->data();

  for (int i=size_ ; i ; --i, image++, image2++)
    *image -= *image2;;

  return *this;
}

intensity &
intensity::operator+=(intensity	*RHS) 
{
  float		*image;
  float		*image2;

  image = data();
  image2 = RHS->data();

  for (int i=size_ ; i ; --i, image++, image2++)
    *image += *image2;

  return *this;
}

intensity &
intensity::operator/=(float	RHS) 
{
  float		*image;

  image = data();

  for (int i=size_ ; i ; --i, image++)
    *image /= RHS;

  return *this;
}

intensity &
intensity::operator*(float	RHS) 
{
  float		*image;

  image = data();

  for (int i=size_ ; i ; --i, image++)
    *image *= RHS;

  return *this;
}

intensity &
intensity::operator/(float	RHS) 
{
  float		*image;

  image = data();

  for (int i=size_; i; --i, image++)
    *image /= RHS;

  return *this;
}

/* normalize image to zero mean and unit variance */
/* only valid for floating point images */
// ImageType *normalize_image(ImageType *im, ImageType *res)

intensity &
intensity::normalize()
{
  int 		i;
  float	 	mean,variance, diff;
  float 	*ip;
  float		min = 9999.0;
  float		max = 0.0;

  if (!valid_)
    return *this;

  /* compute image mean */
  for(mean=0.0,ip=n_image_[0],i=size_;i;--i,++ip)
    mean+= *ip;

  mean /= (float) size_;
  
  /* compute image variance */
  for(variance=0.0,ip=n_image_[0],i=size_;i;--i,++ip)
    {
      diff = *ip - mean;
      variance += diff*diff;
      *ip = diff;
    }
  
  variance = (float) size_ / variance;

  /* scale to unit variance */ 
  for(ip=n_image_[0],i=size_;i;--i,++ip)
    {
      *ip *= variance;
      
      if (max < *ip)
	max = *ip;
      
      if (min > *ip)
	min = *ip;
    }

  max = max - min;
  
  if (max != 0.0)
    for(ip=n_image_[0],i=size_;i;--i,++ip)
      *ip = (*ip - min) / max;


  return *this;
}


intensity &
intensity::umap()
{
  int 		i;
  float	 	mean,variance, diff, sigma;
  float 	*ip;

  if (!valid_)
    return *this;

  /* compute image mean */
  for(mean=0.0,ip=n_image_[0],i=size_;i;--i,++ip)
    mean+= *ip;

  mean /= (float) size_;
  
  /* compute image variance */
  for(variance=0.0,ip=n_image_[0],i=size_;i;--i,++ip)
    {
      diff = *ip - mean;
      variance += diff*diff;
      *ip = diff;
    }
  
  variance /= (float) (size_ - 1 );
  sigma = sqrt(variance);

  /* scale to unit variance */ 
  for(ip=n_image_[0],i=size_;i;--i,++ip) *ip /= sigma;
      
  return *this;
}


intensity &
intensity::kron(intensity	&y)
{
  intensity	*z;
  float		**so_data;
  float		**de_data;

  z =  new intensity(width_ * y.width(), height_ * y.height());

  de_data = (float **)z->data(); 	// get the matrix index
  so_data = (float **)y.data();

  for (int i=0; i < y.height_; i++) 	// Cycle through the source
    for (int j=0; j < y.width_; j++)

      for (int x=0; i < height_; i++) 	// Cycle through the multiplicator
	for (int y=0; j < width_; j++)

	  de_data[i* height_ + x][j* width_ + y]=so_data[i][j] *n_image_[x][y];



  return *z;
}

float
intensity::energy()
{
  float 	energy=0.0;
  float		*fp= n_image_[0];
  
  for (int i=size_; i ; --i, fp++)
    energy += *fp * *fp;

  return (energy);
}


float
intensity::project(intensity	*img)
{
  float		energy=0.0;
  float		*fp = n_image_[0];
  float		*ip = img->data();
  
  for (int i=size_; i ; --i, fp++, ip++)
    energy += *fp * *ip;
  
  return energy;
}

float **
intensity::nr_matrix()
{ 
   int i;

   float * temp=data();
   float ** ret = (float **) malloc(height_*sizeof(float*));
   if (!ret) fprintf(stderr,"Alloc2D: allocation failure 1\n");
   ret -= 1;  /* correct the pointer */

   ret[1] = temp - 1;

   for(i=2;i<=height_;i++) ret[i]=ret[i-1]+width_;

   return ret;
}

// Using the data allocated for this image, generate a new reduced
// image from the image passed in

intensity *
intensity::reduce34(intensity	&RHS)
{

  register int		i,j, t, w,h;
  float			p1 = 2.0 / 3.0;
  float			p2 = 1.0 / 3.0;

  w=RHS.width();
  h=RHS.height();

  // pass through the code horizontally first

  for (j=0; j < h ;j++)
    for (i=0, t=0; i < (w-1) * .75 ; i+=3, t+=4)
      {
	n_image_[j][i]   = RHS[j][t  ]      + RHS[j][t+1] * p2;
	n_image_[j][i+1] = RHS[j][t+1] * p1 + RHS[j][t+2] * p1;
	n_image_[j][i+2] = RHS[j][t+2] * p2 + RHS[j][t+3];
      }
  
  
  // Now make a vertical pass
  
  for (j=0, t=0; t < (h-3); j+=3, t+=4)
    for (int i=0; i < (w-1) * .75 ; i++)
      {
	n_image_[j][i]   = n_image_[t  ][i]      + n_image_[t+1][i] * p2;
	n_image_[j+1][i] = n_image_[t+1][i] * p1 + n_image_[t+2][i] * p1;
	n_image_[j+2][i] = n_image_[t+2][i] * p2 + n_image_[t+3][i];
      }

  // Construct a subimage and return it

  intensity	*ret = new intensity(int(w * .75), int(h * .75), *this);

  return ret;
}

intensity *
intensity::sample( float factor)
{
   int height=(int)(this->height()*factor);
   int width =(int)(this->width()*factor);
   intensity *ret = new intensity(width,height,(float *)NULL);
   register i, j;
   float ** orig_matrix= this->matrix();
   float ** sample_matrix= ret->matrix();

   for(i=0;i<height;i++)
      for(j=0;j<width;j++)
      {
	 sample_matrix[i][j]=orig_matrix
	   [(int)((float)i/factor)][(int)((float)j/factor)];
      }

   return ret;
}

intensity *
intensity::interpolate( float factor)
{
   int im_height=this->height();
   int im_width =this->width();
   int height=(int)(im_height*factor);
   int width =(int)(im_width*factor);
   intensity *ret = new intensity(width,height,(float *)NULL);
   register i, j;
   float ** orig_matrix= this->matrix();
   float ** sample_matrix= ret->matrix();
   float x,y, f1, f2, f3, f4, f;
   int r,rr,c,cc,a,b;

   for(i=0;i<height;i++)
      for(j=0;j<width;j++)
      {
       // do bilinear interpolation
	 x = (float) i/factor;
	 y = (float) j/factor;
	 r = (int) x;
	 c = (int) y;
	 a = x - r;
	 b = y - c;

	// Take into account the limits of the original image
	 rr = (r+1 < im_height) ? r+1 : r;
	 cc = (c+1 < im_width) ? c+1 : c;

	 f1 = orig_matrix[r][c];
	 f2 = orig_matrix[r][cc]; 
	 f3 = orig_matrix[rr][c];
	 f4 = orig_matrix[rr][cc];

	 sample_matrix[i][j]= 
	   (1-a)*(1-b)*f1 + b*(1-a)*f2 + a*(1-b)*f3 + a*b*f4;
      }

   return ret;
}

void 
intensity::save(char * file)
{
   FILE * fd;
   int i,j;

   if ((fd=fopen(file,"w"))==NULL) return;

   for(i=0;i<height_;i++)
   {
      for(j=0;j<width_;j++)
	 fprintf(fd," %g",n_image_[i][j]);
      fprintf(fd,"\n");
   }
   fclose(fd);
}
