
#include "pic.hh"
#include "gif.hh"
#include "intensity.hh"
#include "eigen.hh"
#include "pyramid.hh"

#include <stdio.h>
#include <iostream.h>
#include <unistd.h>
#include <time.h>

extern char*	optarg;


/*
 *  main
 *
 *  The main routine handels the CLI.  We use the optarg routines to
 *  parse out the input file, the output file and the location
 *  of the eigen data files.
 *
 *  We first load the eigens, compute the DFS maps and finally save
 *  the DFS images as gif files. 
 */

// 5-tap filter (Burt88)
float burt_filter[] = {(float)0.0625, (float) 0.25, \
	             (float)0.375,  (float)0.25, (float) 0.0625 };
float baback_filter[] = {(float)0.2, (float) 0.6, (float)0.2 };

//int     pyramid::g_filter_taps = 5;
//float * pyramid::g_filter = gauss_filter(0.5,g_filter_taps);
int     pyramid::g_filter_taps = 3;
float * pyramid::g_filter = baback_filter;

main (int	argc,	char **		argv)
{
  int		c;
  char		eigenbase[50]="heads";
  char		name[50]="input.gif";
  char		output[50]="output";

  while((c=getopt(argc, argv, "hi:d:o:")) !=EOF)
    switch (c)
      {
      case 'i':
	strcpy(name, optarg);
	break;

      case 'd':
	strcpy(eigenbase, optarg);
	break;

      case 'o':
	strcpy(output, optarg);
	break;
      case 'h':
      default:
	
	cout << "USAGE:  " << argv[0] << " -i  <filename> " << endl;
	cout << "        " << argv[0] << " -d  <data directory>" << endl;
	cout << "        " << argv[0] << " -h" << endl;
	
	exit(0);
      }

  pic::verbose();		// Let's see the sizes of the images

  eigen		e_pic(eigenbase); 

// URS: input directly from file (PGM format so far)
  intensity	i_pic(name);	   

  char		buf[100];
  int		i=0;
  gif           **gifs;

  pic::quiet();
  e_pic.load_all();		// load up all eigen bases
      
  pyramid image_pyr(i_pic,0.9);
  pyramid distance_pyr(e_pic,image_pyr); 

// URS: pick up minimum at each scale of the gaussian pyramid
  distance_pyr.found_face(image_pyr);

  i=distance_pyr.levels();
  gifs = new gif*[i];
  while (i-- )
    {
      gifs[i]=new gif(image_pyr.image(i));
      sprintf (buf, "%s%d.gif",output, i);
      gifs[i]->save(buf);
      sprintf (buf, "%s%d.dfs.flt",output,i);
      distance_pyr.image(i)->save(buf);
    }
}
