#include <iostream.h>
#include <stdio.h>
#include <math.h>

#include "convolve.h"  // Baback's convolution code

#include "pyramid.hh"

// pyramid::g_filters = {(0.25-0.4*0.5),0.25, // { -2, -1, 0, 1, 2 }
//		      0.4,
//		      0.25, (0.25-0.4*0.5)};

// pyramid::g_filters = {0.0625, 0.25, // 5-tap filter (Burt88)
//		      0.375,
//		      0.25, 0.0625};

float * 
gauss_filter(float scalefact, int numtaps)
{
// Generates a low-pass gaussian filter for decimation by a factor 'scalefact'
// The number of taps is truncated to 'numtaps'

   int i;
   float x;
   float sum=1;
   int mid=(numtaps-1)/2;
   float * mat=new float[numtaps]; 
   float sigma=0.35/scalefact;
 
   mat[mid]=1.0;
   for(i=-mid;i<0;i++)
   {
      x=i/sigma;
      x*=x/2;
      mat[i+mid]=mat[mid-i]=exp(- x);
      sum+=2*mat[i+mid];
   }
 
   for(i=-mid;i<=mid;i++) mat[i+mid]/=sum;
 
   return mat;
}

pyramid::pyramid(intensity	&image)
{
  // standard implementation: scale 2

  register int 	i;
  float		temp_h;
  float		temp_w;
  intensity * filtered_image_temp;

  temp_h = (float) image.height();
  temp_w = (float) image.width();
  if (temp_h < temp_w)                   // determine # of octaves
    count_ =(int) (log(temp_h) / log(2.0));	
  else
    count_ =(int) (log(temp_w) / log(2.0));	

  images_ = new intensity*[count_]; 	// allocate space for entire array

  images_[0] = new intensity(image);
  
  for(i = 1; i < count_ ; ++i)
  {
      int		 height;
      int		 width;
      float		 val;
      float **    input_image=NULL;
      float **    filtered_image=NULL;

      height = images_[i-1]->height();
      width  = images_[i-1]->width();
      
      filtered_image_temp = new intensity(width,height);

      // convolve with Gaussian kernel (separable filter)
      
      input_image=images_[i-1]->nr_matrix();
      filtered_image=filtered_image_temp->nr_matrix();
      conv2d_sep(input_image, filtered_image, height, width, 
		 g_filter-1, g_filter_taps, g_filter-1, g_filter_taps);

      // Now subsample image
      images_[i] = filtered_image_temp->sample(0.5);

      free(input_image+1);
      free(filtered_image+1);
      delete filtered_image_temp;
      info();
  }
}

pyramid::pyramid(intensity	&image, float scalefact)
{
  register int 	i;
  float		temp_h;
  float		temp_w;
  intensity * image_temp;
  intensity * filtered_image_temp;

  // construct a standard (scale 2) gaussian pyramid for interpolation
  pyramid * model = new pyramid(image);

  temp_h = (float) image.height()/20.0;  // minimum window height = 17
  temp_w = (float) image.width()/10.0;   // minimum window width  = 11;
  if (temp_h < temp_w)                   // determine # of scales
    count_ =(int) (log(temp_h) / log(1.0/scalefact));	
  else
    count_ =(int) (log(temp_w) / log(1.0/scalefact));	

  images_ = new intensity*[count_]; 	// allocate space for entire array

  images_[0] = new intensity(image);
  
  for(i = 1; i < count_ ; ++i)
  {
      float scale = pow (scalefact, i);
      float floatindex = log(scale)/log(0.5);
      int index = (int) (log(scale)/log(0.5));
      if (floatindex != (float) index) index++;
      float int_factor=scale/pow(0.5,index);

      images_[i] = model->images_[index]->interpolate(int_factor);

      info();
  }
  delete model;
}

pyramid::pyramid(eigen e_pic, pyramid inppyr)
{
  int i;
  count_ = inppyr.levels();
  images_ = new intensity*[count_]; 	// allocate space for entire array
  for(i=0;i<count_;i++)
  {
     e_pic.d_f_s(*inppyr.images_[i],10);  // use 10 eigenfaces
     images_[i]=new intensity(*e_pic.dfs());
  }
}

void 
pyramid::found_face(pyramid image_pyr)
{

// NOTE: Some of the data in this function definition are ad-hoc figures
// for debugging purposes. At this stage a minimum of the criterion
// function is picked up at each scale of the gaussian pyramid

   int i;
   char * buf = new char[80];
   for(i=0;i<count_;i++)
   {
      // scan subimages and look for minima
      float * temp=images_[i]->data();
      int rows =images_[i]->height();
      int cols=images_[i]->width();
      int size=rows*cols;
      int minrow=0, mincol=0;

      float minimum=999999999.0;
      for(int j=0;j<size;j++,temp++)
	 if ((*temp)<minimum)
	 {
	    minimum=*temp;
	    minrow=j/cols;
	    mincol=j-minrow*cols;
	 }
      intensity subimage(9,17,*image_pyr.images_[i],mincol,minrow);
      gif face_window(&subimage);
      sprintf (buf, "show%d.gif", i);
      face_window.save(buf);
      cout << i << ", (" << minrow << ", " << mincol << ") " << minimum;
      cout << "\n" ;
   }
}

pyramid::~pyramid()
{

  if (valid_)
    delete[] images_;
}
